import pytest
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm

from sqlalchemy_utils.compat import _select_args
from sqlalchemy_utils.types import json


@pytest.fixture
def Document(Base):
    class Document(Base):
        __tablename__ = 'document'
        id = sa.Column(sa.Integer, primary_key=True)
        json = sa.Column(json.JSONType)
    return Document


@pytest.fixture
def AAA(Base):
    class AAA(Base):
        __tablename__ = 'aaa'
        id = sa.Column(sa.Integer, primary_key=True)
        json_column = sa.Column(json.JSONType, nullable=False)

    return AAA


@pytest.fixture
def BBB(Base, AAA):
    class BBB(Base):
        __tablename__ = 'bbb'
        id = sa.Column(sa.Integer, primary_key=True)
        aaa_id = sa.Column(sa.Integer, sa.ForeignKey('aaa.id'), nullable=True)
        parent = sa_orm.relationship('AAA', backref='children')

    return BBB


@pytest.fixture
def init_models(Document):
    pass


class JSONTestCase:
    def test_list(self, session, Document):
        document = Document(
            json=[1, 2, 3]
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.json == [1, 2, 3]

    def test_parameter_processing(self, session, Document):
        document = Document(
            json={'something': 12}
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.json == {'something': 12}

    def test_non_ascii_chars(self, session, Document):
        document = Document(
            json={'something': 'äääööö'}
        )

        session.add(document)
        session.commit()

        document = session.query(Document).first()
        assert document.json == {'something': 'äääööö'}

    def test_compilation(self, Document, session):
        query = sa.select(*_select_args(Document.json))
        # the type should be cacheable and not throw exception
        session.execute(query)

    def test_unhashable_type(self, AAA, BBB, session):
        """Verify there are no TypeErrors with certain JSON queries.

        This test will fail under these conditions:

        1.  sqlalchemy versions 1.4.19 through 1.4.23 are installed.
        2.  `JSONType.hashable` is not set to False.

        For more info, see:

        *   kvesteri/sqlalchemy-utils#543
        *   sqlalchemy/sqlalchemy#6924
        """

        session.add(AAA(id=13, json_column=[1, 2, 3]))
        session.add(BBB(id=333, aaa_id=13))

        query = session.query(AAA).join(BBB).with_entities(
            AAA.id, AAA.json_column,
        )
        assert len(list(query)) == 1


@pytest.mark.skipif('json.json is None')
@pytest.mark.usefixtures('sqlite_memory_dsn')
class TestSqliteJSONType(JSONTestCase):
    pass


@pytest.mark.skipif('json.json is None')
@pytest.mark.usefixtures('postgresql_dsn')
class TestPostgresJSONType(JSONTestCase):
    pass



import pytest
import sqlalchemy as sa
import sqlalchemy.orm

from sqlalchemy_utils import (  # noqa
    PhoneNumber,
    PhoneNumberParseException,
    PhoneNumberType,
    types
)
from sqlalchemy_utils.compat import _select_args

VALID_PHONE_NUMBERS = (
    "040 1234567",
    "+358 401234567",
    "09 2501234",
    "+358 92501234",
    "0800 939393",
    "09 4243 0456",
    "0600 900 500",
)


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = "user"
        id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
        name = sa.Column(sa.Unicode(255))
        phone_number = sa.Column(PhoneNumberType())

    return User


@pytest.fixture
def init_models(User):
    pass


@pytest.fixture
def phone_number():
    return PhoneNumber("040 1234567", "FI")


@pytest.fixture
def user(session, User, phone_number):
    user = User()
    user.name = "Someone"
    user.phone_number = phone_number
    session.add(user)
    session.commit()
    return user


@pytest.mark.skipif("types.phone_number.phonenumbers is None")
class TestPhoneNumber:
    @pytest.mark.parametrize("raw_number", VALID_PHONE_NUMBERS)
    def test_valid_phone_numbers(self, raw_number):
        number = PhoneNumber(raw_number, "FI")
        assert number.is_valid_number()

    @pytest.mark.parametrize("raw_number", ("abc", "+040 1234567"))
    def test_invalid_phone_numbers__constructor_fails(self, raw_number):
        with pytest.raises(PhoneNumberParseException):
            PhoneNumber(raw_number, "FI")

    @pytest.mark.parametrize("raw_number", ("0111234567", "358"))
    def test_invalid_phone_numbers__is_valid_number(self, raw_number):
        number = PhoneNumber(raw_number, "FI")
        assert not number.is_valid_number()

    def test_invalid_phone_numbers_throw_dont_wrap_exception(self, session, User):
        with pytest.raises(PhoneNumberParseException):
            session.execute(
                User.__table__.insert().values(name="Someone", phone_number="abc")
            )

    def test_phone_number_attributes(self):
        number = PhoneNumber("+358401234567")
        assert number.e164 == "+358401234567"
        assert number.international == "+358 40 1234567"
        assert number.national == "040 1234567"

    def test_phone_number_attributes_for_short_code(self):
        """
        For international and national shortcode remains the same, if we pass
        short code to PhoneNumber library without giving check_region it will
        raise exception
        :return:
        """
        number = PhoneNumber("72404", check_region=False)
        assert number.e164 == "+072404"
        assert number.international == "72404"
        assert number.national == "72404"

    def test_phone_number_str_repr(self):
        number = PhoneNumber("+358401234567")
        assert str(number) == number.national

    def test_phone_number_hash(self):
        number1 = PhoneNumber("+821023456789")
        number2 = PhoneNumber("+82 10-2345-6789")
        assert hash(number1) == hash(number2)
        assert hash(number1) == hash(number1.e164)
        assert {number1} == {number2}


@pytest.mark.skipif("types.phone_number.phonenumbers is None")
class TestPhoneNumberType:
    def test_query_returns_phone_number_object(self, session, User, user, phone_number):
        queried_user = session.query(User).first()
        assert queried_user.phone_number == phone_number

    def test_phone_number_is_stored_as_string(self, session, user):
        result = session.execute(
            sa.text('SELECT phone_number FROM "user" WHERE id=:param'),
            {"param": user.id},
        )
        assert result.first()[0] == "+358401234567"

    def test_phone_number_with_extension(self, session, User):
        user = User(phone_number="555-555-5555 Ext. 555")

        session.add(user)
        session.commit()
        session.refresh(user)
        assert user.phone_number.extension == "555"

    def test_empty_phone_number_is_equiv_to_none(self, session, User):
        user = User(phone_number="")

        session.add(user)
        session.commit()
        session.refresh(user)
        assert user.phone_number is None

    def test_uses_phonenumber_class_as_python_type(self):
        assert PhoneNumberType().python_type is PhoneNumber

    @pytest.mark.usefixtures("user")
    def test_phone_number_is_none(self, session, User):
        phone_number = None
        user = User()
        user.name = "Someone"
        user.phone_number = phone_number
        session.add(user)
        session.commit()
        queried_user = session.query(User)[1]
        assert queried_user.phone_number is None
        result = session.execute(
            sa.text('SELECT phone_number FROM "user" WHERE id=:param'),
            {"param": user.id},
        )
        assert result.first()[0] is None

    def test_scalar_attributes_get_coerced_to_objects(self, User):
        user = User(phone_number="050111222")

        assert isinstance(user.phone_number, PhoneNumber)

    def test_compilation(self, User, session):
        query = sa.select(*_select_args(User.phone_number))
        # the type should be cacheable and not throw exception
        session.execute(query)


@pytest.mark.skipif("types.phone_number.phonenumbers is None")
class TestPhoneNumberComposite:
    @pytest.fixture
    def User(self, Base):
        class User(Base):
            __tablename__ = "user"
            id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)
            name = sa.Column(sa.String(255))
            _phone_number = sa.Column(sa.String(255))
            country = sa.Column(sa.String(255))
            phone_number = sa.orm.composite(PhoneNumber, _phone_number, country)

        return User

    @pytest.fixture
    def user(self, session, User):
        user = User()
        user.name = "Someone"
        user.phone_number = PhoneNumber("+35840111222", "FI")
        session.add(user)
        session.commit()
        return user

    def test_query_returns_phone_number_object(self, session, User, user):
        queried_user = session.query(User).first()
        assert queried_user.phone_number.national == "040 111222"
        assert queried_user.phone_number.region == "FI"



import uuid

import pytest
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from sqlalchemy_utils import UUIDType
from sqlalchemy_utils.compat import _select_args


@pytest.fixture
def User(Base):
    class User(Base):
        __tablename__ = 'user'
        id = sa.Column(UUIDType, default=uuid.uuid4, primary_key=True)

        def __repr__(self):
            return 'User(%r)' % self.id
    return User


@pytest.fixture
def init_models(User):
    pass


class TestUUIDType:
    def test_repr(self):
        plain = UUIDType()
        assert repr(plain) == 'UUIDType()'

        text = UUIDType(binary=False)
        assert repr(text) == 'UUIDType(binary=False)'

        not_native = UUIDType(native=False)
        assert repr(not_native) == 'UUIDType(native=False)'

    def test_commit(self, session, User):
        obj = User()
        obj.id = uuid.uuid4().hex

        session.add(obj)
        session.commit()

        u = session.query(User).one()

        assert u.id == obj.id

    def test_coerce(self, User):
        obj = User()
        obj.id = identifier = uuid.uuid4().hex

        assert isinstance(obj.id, uuid.UUID)
        assert obj.id.hex == identifier

        obj.id = identifier = uuid.uuid4().bytes

        assert isinstance(obj.id, uuid.UUID)
        assert obj.id.bytes == identifier

    def test_compilation(self, User, session):
        query = sa.select(*_select_args(User.id))

        # the type should be cacheable and not throw exception
        session.execute(query)

    def test_literal_bind(self, User):
        expr = (User.id == 'b4e794d6-5750-4844-958c-fa382649719d').compile(
            dialect=postgresql.dialect(),
            compile_kwargs={'literal_binds': True}
        )
        assert str(expr) == (
            '''"user".id = \'b4e794d6-5750-4844-958c-fa382649719d\''''
        )



import pytest
import sqlalchemy as sa

from sqlalchemy_utils import get_fk_constraint_for_columns, has_index


class TestHasIndex:

    @pytest.fixture
    def table(self, Base):
        class ArticleTranslation(Base):
            __tablename__ = 'article_translation'
            id = sa.Column(sa.Integer, primary_key=True)
            locale = sa.Column(sa.String(10), primary_key=True)
            title = sa.Column(sa.String(100))
            is_published = sa.Column(sa.Boolean, index=True)
            is_deleted = sa.Column(sa.Boolean)
            is_archived = sa.Column(sa.Boolean)

            __table_args__ = (
                sa.Index('my_index', is_deleted, is_archived),
            )
        return ArticleTranslation.__table__

    def test_column_that_belongs_to_an_alias(self, table):
        alias = sa.orm.aliased(table)
        with pytest.raises(TypeError):
            assert has_index(alias.c.id)

    def test_compound_primary_key(self, table):
        assert has_index(table.c.id)
        assert not has_index(table.c.locale)

    def test_single_column_index(self, table):
        assert has_index(table.c.is_published)

    def test_compound_column_index(self, table):
        assert has_index(table.c.is_deleted)
        assert not has_index(table.c.is_archived)

    def test_table_without_primary_key(self):
        article = sa.Table(
            'article',
            sa.MetaData(),
            sa.Column('name', sa.String)
        )
        assert not has_index(article.c.name)


class TestHasIndexWithFKConstraint:
    def test_composite_fk_without_index(self, Base):

        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
            )

        table = Article.__table__
        constraint = get_fk_constraint_for_columns(
            table,
            table.c.author_first_name,
            table.c.author_last_name
        )
        assert not has_index(constraint)

    def test_composite_fk_with_index(self, Base):

        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
                sa.Index(
                    'my_index', author_first_name, author_last_name
                )
            )

        table = Article.__table__
        constraint = get_fk_constraint_for_columns(
            table,
            table.c.author_first_name,
            table.c.author_last_name
        )
        assert has_index(constraint)

    def test_composite_fk_with_partial_index_match(self, Base):

        class User(Base):
            __tablename__ = 'user'
            first_name = sa.Column(sa.Unicode(255), primary_key=True)
            last_name = sa.Column(sa.Unicode(255), primary_key=True)

        class Article(Base):
            __tablename__ = 'article'
            id = sa.Column(sa.Integer, primary_key=True)
            author_first_name = sa.Column(sa.Unicode(255))
            author_last_name = sa.Column(sa.Unicode(255))
            __table_args__ = (
                sa.ForeignKeyConstraint(
                    [author_first_name, author_last_name],
                    [User.first_name, User.last_name]
                ),
                sa.Index(
                    'my_index',
                    author_first_name,
                    author_last_name,
                    id
                )
            )

        table = Article.__table__
        constraint = get_fk_constraint_for_columns(
            table,
            table.c.author_first_name,
            table.c.author_last_name
        )
        assert has_index(constraint)



import pytest

import sqlalchemy_utils.compat


@pytest.mark.parametrize(
    "version, expected",
    (
        ("2.0.0b3", (2, 0, 0)),
        ("2.0.0b", (2, 0, 0)),
        ("2.0.0", (2, 0, 0)),
        ("2.0.", (2, 0)),
        ("2.0", (2, 0)),
        ("2.", (2,)),
        ("2", (2,)),
        ("", ()),
    ),
)
def test_get_sqlalchemy_version(version, expected):
    assert sqlalchemy_utils.compat.get_sqlalchemy_version(version) == expected
